#pragma glslify: import('../common.glsl')

// 2D rotation.
mat2 rot(float th){
    float cs = cos(th), si = sin(th);
    return mat2(cs, -si, si, cs);
}

// Standard hash algorithm that you'll see all over the place. Range: [-1, 1].
vec3 hash33(vec3 p) {
    float n = sin(dot(p, vec3(7.0, 157., 113.)));
    return fract(vec3(64.0*n, 8.0*n, n)*32768.0)*2.-1.;
}

float map(vec3 p, out vec4 color) {
    vec3 center = vec3(0.5);
    vec3 pp = fract(p) - center;
    float angle = mod(atan(pp.y, pp.x) / TWO_PI + 0.25 + u_time/5.0, 1.);
    float vol = texture2D(u_frequencies, vec2(angle, 0)).x;
    color = getTweenedTexture(u_palette, vol*2.);

    if (mod(floor(p), 2.) == vec3(0)) {
      return length(pp) - vol/1.1 - 0.05;
    }

    return length(pp) - 0.03;
}

const float eps = 0.005;

vec3 getNormal(in vec3 p, out vec4 color) {
	return normalize(vec3(
		map(vec3(p.x+eps,p.y,p.z), color)-map(vec3(p.x-eps,p.y,p.z), color),
		map(vec3(p.x,p.y+eps,p.z), color)-map(vec3(p.x,p.y-eps,p.z), color),
		map(vec3(p.x,p.y,p.z+eps), color)-map(vec3(p.x,p.y,p.z-eps), color)
	));
}

vec3 raymarch(vec3 ro, vec3 rd) {
	float t = 0.0;
	vec3 col = vec3(0.);

    int finalIndex = -1;
    vec3 finalPos;
    float finalDistance = -1.;

    for (int i = 0; i < 56; i++) {
        vec3 pos = ro + rd*t;
        vec4 color;
        float d = map(pos, color);

        if (d<0.005 || t>20.) {
          finalIndex = i;
          finalPos = pos;
          finalDistance = d;
          break;
        }

        float td = min(0.25, d);
        t += td * 0.5;

        float atten = max(1.-0.2/max(d, 0.001), 0.)*0.3;
        //float atten = min(0.2/max(d, 0.001), 1.)*0.15;

        // Adds a bit of jitter.
        pos += hash33(pos)*0.03;

        // Compute background color.
        vec3 texCol = getTweenedTexture(u_palette, abs(sin((pos.xy + pos.zy + pos.zx)/4.0)).x/1.7).xyz / 4.0;
        col += texCol*atten;

        // A bit of ray warping.
        mat2 rt = rot(0.03*td);
        rd.xy *= rt;
        rd.zx *= rt;
    }

    if (finalIndex != -1) {
      // Use trivial occlusion to shade.
      vec4 color;
      vec3 surfNormal = getNormal(finalPos, color);
      col += 1. * color.rgb * surfNormal.y;
    }

    // Return the final color.
    return col;
}

void main() {
    // Screen coordinates.
	vec2 uv = (gl_FragCoord.xy - u_resolution.xy*0.5) / u_resolution.y;

    // Normalized directional ray. The line directly below gives a mild
    // fish lens effect, and the one below it is the non-warped version.
    // Use whichever one you prefer.
    vec3 rd = normalize(vec3(uv, sqrt(1.-dot(uv, uv)*0.5)*0.5));
    //vec3 rd = normalize(vec3(uv, 0.5));

    // Rotate the directional ray to give the observer a look around.
    // I think of it as a cheap, lazy-man's camera. :)
    rd.xy *= rot(u_time*0.3);
    rd.xz *= rot(u_time*0.15); // Extra variance.

    // Our startoff postion, which moves forward in the Z-direction.
    vec3 ro = vec3(0.0, 0.0, u_time*1.5);

    // Calculate the color.
    vec3 sCol = raymarch(ro, rd);

    // Put it on the screen. Too easy. :)
	gl_FragColor = vec4(clamp(sCol, 0., 1.), 1.0);

    // Now give everything a really vignetted overlay.
    uv = gl_FragCoord.xy / u_resolution.xy;
    uv *= 1.0 - uv.yx;
    float vig = pow(uv.x*uv.y * 50.0, 0.2);
    gl_FragColor.rgb *= vig;
}
