#pragma glslify: import('../common.glsl')

float hash(float n)
{
  highp float a  = 12.9898;
  highp float b  = 78.233;
  highp float c  = 43758.5453;
  highp float dt = dot(vec2(n, n) ,vec2(a,b));
  highp float sn = mod(dt, 3.14);
  return fract(sin(sn) * c);
}

vec3 ClosestPoint(vec3 ro, vec3 rd, vec3 p) {
    return ro + max(0., dot(p-ro, rd))*rd;
}

vec3 ClosestPointFromLineToLine(vec3 l1Start, vec3 l1End, vec3 l2Start, vec3 l2End, out float dist)
{
  vec3  u = l1End - l1Start;
  vec3  v = l2End - l1Start;
  vec3  w = l1Start - l2Start;
  float a = dot(u, u);
  float b = dot(u, v);
  float c = dot(v, v);
  float d = dot(u, w);
  float e = dot(v, w);
  float D = a*c - b*b;
  float sc, tc;

  if (D < 0.001) {
    sc = 0.0;
    tc = (b>c ? d/b : e/c);
  } else {
    sc = (b*e - c*d) / D;
    tc = (a*e - b*d) / D;
  }

  vec3 closestPoint = (tc * v);
  vec3 dP = w + (sc * u) - closestPoint;
  dist = length(dP);

  // Make sure we're not following the line the wrong direction.
  if (dot(closestPoint - l2Start, l2End - l2Start) < 0.) {
    dist = 1000.;
  }

  return closestPoint;
}

float bokehRing(vec3 ro, vec3 rd, vec3 p, float size, float blur) {
  float dist = length(cross(p-ro, rd));
  float m = smoothstep(size, size*(1.-blur), dist);
  m *= mix(.25, 1., smoothstep(.95*size, size, dist));
  return m;
}

vec3 disc(vec3 cameraPos, vec3 rd, vec3 p) {
  float d = length(p-cameraPos);
  float distanceFromCamera = length(cameraPos - p);
  float distanceFromOrigin = length(p);

  float size = .55 * smoothstep(0.1, 6.1, distanceFromOrigin);
  vec3 rp = ClosestPoint(cameraPos, rd, p);

  // Focus is on a plane from origin.
  float distanceFromFocalPlane = abs(15. - length(p - cameraPos));
  float blur = (distanceFromFocalPlane < 4.) ? 0.05 : min(distanceFromFocalPlane/2., .5);
  float m = bokehRing(cameraPos, rd, p, .55, blur);

  // Fade out as we get closer to origin.
  float brightness = min(distanceFromOrigin/1.7, 3.);
  size = min(distanceFromOrigin, size);

  return m * brightness * getTweenedTexture(u_palette, length(p - vec3(0, 0, 0))/7.0).rgb;
}

vec3 line(vec3 cameraPos, vec3 rayDirection, vec3 lineOrigin, vec3 lineDirection, vec3 color) {
  float dist;
  vec3 closestPoint = ClosestPointFromLineToLine(cameraPos, cameraPos + rayDirection * 100., lineOrigin, lineOrigin + lineDirection * 100., dist);
  float distFromOrigin = length(closestPoint - vec3(0));
  float distFromCamera = length(cameraPos - closestPoint);

  return color * (1. - smoothstep(0., 0.03 + distFromOrigin/7., dist)) / (1.5 * distFromCamera/2.0);
}

vec3 CreateRay(vec2 uvOrig, vec3 cameraPos, vec3 lookAt) {
  vec2 uv = uvOrig;
  uv -= .5;
  uv.x *= u_resolution.x/u_resolution.y;

  vec3 rayOrigin = cameraPos;
  vec3 forward = normalize(lookAt - rayOrigin);
  vec3 right = cross(vec3(0, 1, 0), forward);
  vec3 up = cross(forward, right);
  vec3 center = rayOrigin + forward;
  vec3 intersect = center + (uv.x * right) + (uv.y * up);

  return normalize(intersect - rayOrigin);
}

#define BALL_SETS 3.
#define NUM_RAYS  12.

void main()
{
  // Create normalized coord system centered on 0,0
  vec2 uv = gl_FragCoord.xy / u_resolution.xy;

  vec3 cameraPos = vec3(sin(u_time)*10., 0., -10.*cos(u_time));
  vec3 lookAt = vec3(0, 0, 0);
  vec3 ray = CreateRay(uv, cameraPos, lookAt);
  vec3 col = vec3(0.);

  // Render the discs.
  for (float j = 0.; j < 24. * BALL_SETS; j += 1.) {
    float i = mod(j, 24.);
    float zeroToOne = i / 23.;
    float vol = getFrequencyValue(zeroToOne, FREQ_AMPLITUDE);
    vec3 ballRay = normalize(vec3(hash(j)-0.5, hash(j*3.)-0.5, hash(j*5.)-0.5));

    col += disc(cameraPos, ray, ballRay * vol * 15.);
  }

  // Render the rays.
  for (float j = 0.; j < NUM_RAYS; j += 1.) {
    float vol = getFrequencyValue(j/(NUM_RAYS-1.), FREQ_AMPLITUDE);
    vec3 color = getTweenedTexture(u_palette, vol + 0.1).rgb;
    vol /= 4.;

    // Pick a semi-random direction.
    float random = hash(j);
    vec3 lightDir = normalize(vec3(sin(j+random+vol)+.1, cos(j+vol) + cos(u_time/100.)/3. -.1, random));

    // Turn down brightness if it's pointing towards us, avoids flashes.
    float dotRay = dot(lightDir, ray);
    float brightness = 1. - smoothstep(-0.1, -0.9, dotRay);

    // Offset the origin just a tiny bit in the right direction so they don't overlap on origin.
    vec3 lightOrigin = vec3(0.) + 0.01 * lightDir;
    col += brightness * line(cameraPos, ray, lightOrigin, lightDir, color) * 1.5;
  }

  gl_FragColor = vec4(col, length(col));
}
