#pragma glslify: import('../common.glsl')

const float eps = 0.005;
const int maxIterations = 128;
const float stepScale = 0.5;
const float stopThreshold = 0.005;

float sphere(in vec3 p, in vec3 centerPos, float radius) {
	return length(p-centerPos) - radius;
}

float sinusoidBumps(in vec3 p){
    float samplePoint = mod((atan(p.x + 0.5, p.z) + PI)/PI/2., 1.0);
    //float samplePoint = clamp(length(p), 0., 1.);
    float vol = getFrequencyValueBilinear(samplePoint, FREQ_AMPLITUDE) + 0.03;
    return -(vol * 8. * (0.5 + sin(p.x*16.+u_time*0.57)*cos(p.y*16.+u_time*2.17)*sin(p.z*16.-u_time*1.31) + 0.5*sin(p.x*32.+u_time*0.07)*cos(p.y*32.+u_time*2.11)*sin(p.z*32.-u_time*1.23)));
}

float scene(in vec3 p) {
	return sphere(p, vec3(0., 0. , 3.), 2.4) + 0.04*sinusoidBumps(p);
}

vec3 getNormal(in vec3 p) {
    float ref = scene(p);

	// 6-tap normalization. Probably the most accurate, but a bit of a cycle waster.
    return normalize(vec3(
      scene(vec3(p.x+eps,p.y,p.z))-scene(vec3(p.x-eps,p.y,p.z)),
      scene(vec3(p.x,p.y+eps,p.z))-scene(vec3(p.x,p.y-eps,p.z)),
      scene(vec3(p.x,p.y,p.z+eps))-scene(vec3(p.x,p.y,p.z-eps))
    ));
}

float rayMarching( vec3 origin, vec3 dir, float start, float end ) {

	float sceneDist = 1e4;
	float rayDepth = start;
	for (int i = 0; i < maxIterations; i++) {
		sceneDist = scene(origin + dir * rayDepth);
		if ((sceneDist < stopThreshold) || (rayDepth >= end)) {
			break;
		}
		rayDepth += sceneDist * stepScale;
	}

	if (sceneDist >= stopThreshold)
	  rayDepth = end;
	else
	  rayDepth += sceneDist;

	return rayDepth;
}


void main(void) {
    vec2 aspect = vec2(u_resolution.x/u_resolution.y, 1.0);
	vec2 screenCoords = (2.0*gl_FragCoord.xy/u_resolution.xy - 1.0)*aspect;

	vec3 lookAt = vec3(0.,0.,0.);
	vec3 camPos = vec3(0., 0., -0.8);

    // Camera setup.
    vec3 forward = normalize(lookAt-camPos);
    vec3 right = normalize(vec3(forward.z, 0., -forward.x));
    vec3 up = normalize(cross(forward, right));
    float FOV = 0.5;

    // ro - Ray origin. Every ray starts from this point, then is cast in the rd direction.
    // rd - Ray direction. This is our one-unit-long direction ray.
    vec3 ro = camPos;
    vec3 rd = normalize(forward + FOV*screenCoords.x*right + FOV*screenCoords.y*up);

	// The screen's background color.
    vec3 bgcolor = vec3(0, 0, 0);

	// Ray marching.
	const float clipNear = 0.0;
	const float clipFar = 10.0;
	float dist = rayMarching(ro, rd, clipNear, clipFar ); // See the function itself for an explanation.
	if (dist >= clipFar) {
	    gl_FragColor = vec4(bgcolor, 0.);
	    return;
	}

	// sp - Surface position.
	vec3 sp = ro + rd*dist;

	// Anyway, check out the function itself.
	vec3 surfNormal = getNormal(sp);

	// lp - Light position.
	vec3 lp = vec3(0.5, 0.5, -1.);

	// ld - Light direction. The point light direction goes from the light's position to the surface point we've hit on the sphere. I haven't
	// normalized it yet, because I'd like to take the length first, but it will be.
	vec3 ld = lp-sp;
	vec3 lightColor = vec3(1.,0.97,0.92);

	float len = length( ld ); // Distance from the light to the surface point.
	ld /= len; // Normalizing the light-to-surface, aka light-direction, vector.
	float lightAtten = min( 1.0 / ( 0.25*len*len ), 1.0 ); // Keeps things between 0 and 1.
	vec3 ref = reflect(-ld, surfNormal);

    // Start with black. If we had global ambient lighting, then I guess we could add it here, or later. It's a preference thing.
	vec3 sceneColor = vec3(1) * surfNormal.y * 0.1;

	float bumps =  sinusoidBumps(sp);
	vec3 objColor = getTweenedTexture(u_palette, clamp(abs(bumps / 4.0), 0., 1.)).rgb;

	float ambient = .5;
	float specularPower = 75.0;
	float diffuse = max(0.0, dot(surfNormal, ld));
	float specular = max(0.0, dot(ref, normalize(camPos-sp)));
	specular = pow(specular, specularPower);

	// Bringing all the lighting components together to color the screen pixel.
	sceneColor += (objColor*(diffuse*0.8+ambient)+specular*0.4)*lightColor*lightAtten;
    sceneColor = clamp(sceneColor, 0.0, 1.0);

	//gl_FragColor = sqrt(vec4(sceneColor, .75));
	//gl_FragColor = vec4(pow(sceneColor, vec3(1.0/2.0)), 0.8);
	gl_FragColor = vec4(sceneColor, sqrt(length(sceneColor)));
}
